;; Copyright (C) 2016 Antonin Houska
;;
;; This file is part of PGQA.
;;
;; PGQA is free software: you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later
;; version.

;; PGQA is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License qalong
;; with PGQA. If not, see <http://www.gnu.org/licenses/>.

(require 'eieio)

(defclass pgqa-node ()
  (
   (region
    :initarg :region
    :documentation "Start and end position of a nonterminal."
    )

   (markers
    :initarg :markers
    :documentation "Start and end position in the form of a marker."
    )
   )

  "Class representing a generic node of SQL query tree.")

;; TODO Recursion.
(defmethod pgqa-set-markers ((node pgqa-node))
  "Turn `region' into a marker and recurse into children."
  (let* ((reg-vec (oref node region))
	 (reg-start (elt reg-vec 0))
	 (reg-end (elt reg-vec 1))
	 (m-start (make-marker))
	 (m-end (make-marker)))
    (set-marker m-start reg-start)
    (set-marker m-end reg-end)

    ;; The insertion types are such that the start and end markers always span
    ;; only the original region.
    (set-marker-insertion-type m-start t)
    (set-marker-insertion-type m-end nil)

    (oset node markers (vector m-start m-end)))
  )

;; `state' is an instance of `pgqa-deparse-state' class.
;;
;; `indent' determines indentation of the node.
;;
;; `space' indicates that that the node should put space in front of the
;; actual string, unless it starts right after line break. (Caller shouldn't
;; be responsible for the space because he cannot simply determine whether the
;; first token of the node is short enough to fit on the current line. And we
;; don't whant whitespace at the end of line.)
(defmethod pgqa-dump ((node pgqa-node) state indent space)
  "Turn a node and its children into string."
  nil)

;; XXX If adding a new field, check if it's subject to backup / restore by
;; pgqa-dump method of pgqa-operator class. (Is this backup / restore worth
;; separate functions?)
(defclass pgqa-deparse-state ()
  (
   (indent
    :initarg :indent
    ;; This is useful when the query is not printed out structured, i.e. line
    ;; breaks are only enforced by fill-column. In this case the indentation
    ;; passed to pgqa-dump is useless because it gets increased due to
    ;; recursive calls.
    :documentation "Indentation level of the top level query.")

   (next-column
    :initarg :next-column
    :documentation "Column at which the following node (or its leading space)
should start.")

   (no-init-newline
    :initarg :no-init-newline
    :documentation "Do not add line delimiter in front of the whole query even
if `pgqa-multiline-query' is t.")

   (result
    :initarg :result
    :documentation "String to which each node appends its textual
representation.")
   )

  "State of the deparsing process."
  )

(defmethod pgqa-deparse-newline ((state pgqa-deparse-state) indent)
  "Adjust deparse state so that deparsing continues at a new line, properly
indented."

  ;; No recursive increments of indentation if the query should look "flat".
  (unless pgqa-multiline-query
    (setq indent (oref state indent)))

  (let* ((indent-width (* indent tab-width))
	 (result (oref state result)))
    (setq result (concat result "\n"))
    (setq result (concat result
			 (make-string indent-width (string-to-char " "))))
    (oset state result result)
    (oset state next-column indent-width))
  )

(defmethod pgqa-deparse-space ((state pgqa-deparse-state))
  "Write space to deparsing output."
  (oset state result (concat (oref state result) " "))
  (oset state next-column (1+ (oref state next-column))))

;; Prepare for insertion of `str', i.e. add line break or space, as
;; needed. `space' argument has the same semantics as that of `pgqa-dump'
;; method of `pgqa-node' class and its children.
(defmethod pgqa-deparse-string-prepare ((state pgqa-deparse-state) str indent
					space)
  (let ((col-incr 0))
    (if space
	(setq col-incr (1+ col-incr)))
    (setq col-incr (+ col-incr (string-width str)))

    (if (and fill-column
	     (> (+ (oref state next-column) col-incr) fill-column))
	(pgqa-deparse-newline state indent)
      (if space
	  (pgqa-deparse-space state)))
    )
  )

(defmethod pgqa-deparse-string ((state pgqa-deparse-state) str indent space)
  "Write arbitrary string to deparsing output."

  ;; Decide whether line should be broken first, and do so.
  ;;
  ;; Some characters should not star the new line, so let the related node
  ;; (operator) handle the overfow. Not sure if more sofisticated logic is
  ;; needed. If we e.g. tried to put newline in front of the opening
  ;; parenthesis, it can be nested in another pair of parentheses that we
  ;; break anyway.
  ;;
  ;; TODO Add "(" to the list when parser can handle functions. (It's only
  ;; important for functions to keep "(" stuck to the function name.)
  (unless (or (string= str ")") (string= str "[") (string= str "]"))
    (pgqa-deparse-string-prepare state str indent space))

  (oset state result (concat (oref state result) str))
  (oset state next-column (+ (oref state next-column) (string-width str))))


;; init-col-src is column (in addition to the indentation) at which the source
;; query starts.
(defun pgqa-init-deparse-state (indent init-col-src no-init-newline)
  "Initialize instance of `pgqa-deparse-state'."

  (let ((indent-width (* indent tab-width)))
    (make-instance 'pgqa-deparse-state
		   :indent indent
		   :next-column (+ indent-width init-col-src)
		   :no-init-newline no-init-newline
		   :result (make-string indent-width (string-to-char " ")))))

(defclass pgqa-query (pgqa-node)
  (
   (target-expr :initarg :target-expr)
   (from-expr :initarg :from-expr)
   )
  "A generic SQL query (or subquery).")

(defmethod pgqa-dump ((node pgqa-query) state indent space)
  "Turn query into a string."

  (let ((no-init-newline (oref state no-init-newline)))
    ;; Whatever the current value, no-init-newline is not applicable to
    ;; subqueries.
    (oset state no-init-newline nil)

    (if (slot-boundp node 'target-expr)
	(let ((te (oref node target-expr)))
	  (if (and pgqa-multiline-query (null no-init-newline))
	      (pgqa-deparse-newline state indent))
	  (pgqa-deparse-string state "SELECT" indent space)
	  (pgqa-dump (oref te target-list)  state (1+ indent) t)))

    (if (slot-boundp node 'from-expr)
	(pgqa-dump (oref node from-expr) state indent t))
    )
)

(defclass pgqa-targ-expr (pgqa-node)
  (
   (target-list :initarg :target-list)
   )
  "The target expression"
)

(defmethod pgqa-dump ((node pgqa-targ-expr) state indent space)
  )

(defclass pgqa-from-expr (pgqa-node)
  (
   (from-list :initarg :from-list)
   (qual :initarg :qual)
   )
  "FROM expression of an SQL query."
)

(defmethod pgqa-dump ((node pgqa-from-expr) state indent space)
  (if pgqa-multiline-query
      (progn
	(pgqa-deparse-newline state indent)
	(setq space nil)))
  (pgqa-deparse-string state "FROM" indent space)

  (let ((from-list (oref node from-list)))
    (if (= (length from-list) 1)
	(pgqa-dump (car from-list) state (1+ indent) t)
      ;; Line breaks and indentation are best handled if we turn the list into
      ;; a comma operator.
      ;;
      ;; XXX Shouldn't this be done by parser?
      (let (
	    (l (make-instance 'pgqa-operator :op ","
			      :args from-list
			      :prec pgqa-precedence-comma))
	    )
	(pgqa-dump l state (1+ indent) t)
	)
      )
    )

  (if (slot-boundp node 'qual)
      (progn
	(if pgqa-multiline-query
	    (pgqa-deparse-newline state indent))
	;; `space' should be up-to-date as the FROM clause is mandatory.
	(pgqa-deparse-string state "WHERE" indent space)
	(pgqa-dump (oref node qual) state (1+ indent) t)))
    )

(defclass pgqa-from-list-entry (pgqa-node)
  (
   (obj :initarg :obj)
   (alias :initarg :alias)
   )
  "From list entry (table, join, subquery, ...)"
)

(defmethod pgqa-dump ((node pgqa-from-list-entry) state indent space)
  "Print out FROM list entry (table, join, subquery, etc.)."

  (pgqa-dump (oref node obj) state indent space)

  (if (slot-boundp node 'alias)
      (progn
	(pgqa-deparse-string state "AS" indent t)
	(pgqa-deparse-string state (oref node alias) indent t))
    )
  )

;; Number is currently stored as a string - should this be changed?
(defclass pgqa-number (pgqa-node)
  (
   (value :initarg :value)
   )
  "A number.")

(defmethod pgqa-dump ((node pgqa-number) state indent space)
  "Turn number into a string."
  (let ((str (oref node value)))
    (pgqa-deparse-string state str indent space))
  )

(defclass pgqa-obj (pgqa-node)
  (
   ;; x.y expression can represent either column "y" of table "x" or table "y"
   ;; of schema "x". Instead of teaching parser to recognize the context (is
   ;; it possible?) let's postpone resolution till analysis phase.
   ;;
   ;; Note that the number of arguments is not checked during "raw parsing",
   ;; and that asterisk can be at any position, not only the last one.
   (args :initarg :args)

   (alias :initarg :alias)
   ;; TODO
   ;;
   ;; 1. Remove double-quotes (and remember there were some?)
   ;;
   ;; 2. Turn unquoted values to lower case.
   )
  "Table or column reference.")

(defmethod pgqa-dump ((node pgqa-obj) state indent space)
  "Turn an SQL object into a string."
  (let ((str (mapconcat 'format (oref node args) ".")))
    (pgqa-deparse-string state str indent space)))

(defclass pgqa-expr (pgqa-node)
  (
   (args :initarg :args)
   )
  "A node representing an operation on one or multiple arguments.")


(defclass pgqa-operator (pgqa-expr)
  (
   (op :initarg :op)
   (prec :initarg :prec
	 :documentation "Operator precedence, for the sake of printing.")
   )
  "Generic operator.")

;; TODO Some other constructs need parentheses, e.g. query in a
;; sublink. Should they be given certain precedence just because of this?
(defun pgqa-child-needs-parens (node arg)
  "Find out if argument of an operator node should be parenthesized."
  (let ((prec (oref node prec))
	(prec-child
	 (if (object-of-class-p arg pgqa-operator)
	     (oref arg prec))))
    (and prec prec-child (> prec prec-child)))
  )

(defmethod pgqa-dump ((node pgqa-operator) state indent space)
  "Turn operator expression into a string."
  (let* ((args (oref node args))
	 (nargs (length args))
	 (op (oref node op))
	 (i 0))

    (dolist (arg args)
      (let ((is-unary
	     (null
	      (or (cdr args)
		  (string= op ","))))
	    (parens (pgqa-child-needs-parens node arg))
	    (str)
	    (result-backup)
	    (next-column-backup))

	;; TODO Currently all the unary operators are prefixes. Enhance
	;; when needed.
	(if (or is-unary (> i 0))
	    (let ((space-local space))
	      ;; Never put space in front of comma.
	      (if (string= op ",")
		  (setq space-local nil))
	      (pgqa-deparse-string state op indent space-local)))

	;; If space omission was required by caller, we could have it already
	;; applied.
	(if is-unary
	    (setq space t))

	(if parens
	    (progn
	      (setq str "(")
	      (pgqa-deparse-string state str indent space)
	      ;; No space, whatever follows "(".
	      (setq space nil)))

	;; Comma needs special treatment because it doesn't look nice if it's
	;; the first character on a line.
	(if (string= op ",")
	    (progn
	      (setq result-backup (oref state result))
	      (setq next-column-backup (oref state next-column)))
	  )

	;; Serialize the argument now.
	(pgqa-dump arg state indent space)

	(if
	    ;; If the argument should be followed by comma, line should have
	    ;; broken in front of the argument. So restore the previous state
	    ;; and dump the argument again, with fill-column temporarily
	    ;; decreased by one. That should make the argument appear on the
	    ;; new line too.
	    (and
	     (string= op ",")
	     (>= (oref state next-column) fill-column)
	     (< i (1- nargs)))
	    (progn
	      (oset state result result-backup)
	      (oset state next-column next-column-backup)

	      (let ((fill-column (1- fill-column)))
		(pgqa-dump arg state indent space)
		)
	      )
	  )

	(if parens
	    ;; Never space in front of ")".
	    (pgqa-deparse-string state ")" indent nil))

	(setq i (1+ i))

	;; Space should not be omitted in front of the following operator
	;; string (unless it's comma, see above).
	(setq space t))
      )
    )
)

(defclass pgqa-target-entry (pgqa-node)
  (
   (expr :initarg :expr)
   (alias :initarg :alias)
   )
  "Target list entry")

(defmethod pgqa-dump ((node pgqa-target-entry) state indent space)
  "Turn target list entry into a string."

  (pgqa-dump (oref node expr) state indent space)

  (if (slot-boundp node 'alias)
      (progn
	(pgqa-deparse-string state "AS" indent space)
	(pgqa-deparse-string state (oref node alias) indent space))
    )
)

(provide 'pgqa-node)
